/* $Id: gl.h,v 3.19 1998/07/31 03:23:36 brianp Exp $ */

/* Amicall's for StormC-PPC */
#ifndef STORMPROTOS_GLE_SPROTOS_H
#define STORMPROTOS_GLE_SPROTOS_H

extern "library=gleppcBase" {

extern int gleGetJoinStyle (void);
extern void gleSetJoinStyle (int style);        /* bitwise OR of flags */
extern int gleGetNumSlices(void);
extern void gleSetNumSlices(int slices);

/* draw polyclinder, specified as a polyline */
extern void glePolyCylinder (int npoints,       /* num points in polyline */
		   gleDouble point_array[][3],  /* polyline vertces */
		   float color_array[][3],      /* colors at polyline verts */
		   gleDouble radius);           /* radius of polycylinder */

/* draw polycone, specified as a polyline with radii */
extern void glePolyCone (int npoints,    /* numpoints in poly-line */
		   gleDouble point_array[][3],  /* polyline vertices */
		   float color_array[][3],      /* colors at polyline verts */
		   gleDouble radius_array[]); /* cone radii at polyline verts */

/* extrude arbitrary 2D contour along arbitrary 3D path */
extern void gleExtrusion (int ncp,         /* number of contour points */
		gleDouble contour[][2],     /* 2D contour */
		gleDouble cont_normal[][2], /* 2D contour normals */
		gleDouble up[3],            /* up vector for contour */
		int npoints,            /* numpoints in poly-line */
		gleDouble point_array[][3], /* polyline vertices */
		float color_array[][3]); /* colors at polyline verts */

/* extrude 2D contour, specifying local rotations (twists) */
extern void gleTwistExtrusion (int ncp,         /* number of contour points */
		gleDouble contour[][2],    /* 2D contour */
		gleDouble cont_normal[][2], /* 2D contour normals */
		gleDouble up[3],           /* up vector for contour */
		int npoints,           /* numpoints in poly-line */
		gleDouble point_array[][3],        /* polyline vertices */
		float color_array[][3],        /* color at polyline verts */
		gleDouble twist_array[]);   /* countour twists (in degrees) */

/* extrude 2D contour, specifying local affine tranformations */
extern void gleSuperExtrusion (int ncp,  /* number of contour points */
		gleDouble contour[][2],    /* 2D contour */
		gleDouble cont_normal[][2], /* 2D contour normals */
		gleDouble up[3],           /* up vector for contour */
		int npoints,           /* numpoints in poly-line */
		gleDouble point_array[][3],        /* polyline vertices */
		float color_array[][3],        /* color at polyline verts */
		gleDouble xform_array[][2][3]);   /* 2D contour xforms */

/* spiral moves contour along helical path by parallel transport */
extern void gleSpiral (int ncp,        /* number of contour points */
	     gleDouble contour[][2],    /* 2D contour */
	     gleDouble cont_normal[][2], /* 2D contour normals */
	     gleDouble up[3],           /* up vector for contour */
	     gleDouble startRadius,     /* spiral starts in x-y plane */
	     gleDouble drdTheta,        /* change in radius per revolution */
	     gleDouble startZ,          /* starting z value */
	     gleDouble dzdTheta,        /* change in Z per revolution */
	     gleDouble startXform[2][3], /* starting contour affine xform */
	     gleDouble dXformdTheta[2][3], /* tangent change xform per revoln */
	     gleDouble startTheta,      /* start angle in x-y plane */
	     gleDouble sweepTheta);     /* degrees to spiral around */

/* lathe moves contour along helical path by helically shearing 3D space */
extern void gleLathe (int ncp,        /* number of contour points */
	     gleDouble contour[][2],    /* 2D contour */
	     gleDouble cont_normal[][2], /* 2D contour normals */
	     gleDouble up[3],           /* up vector for contour */
	     gleDouble startRadius,     /* spiral starts in x-y plane */
	     gleDouble drdTheta,        /* change in radius per revolution */
	     gleDouble startZ,          /* starting z value */
	     gleDouble dzdTheta,        /* change in Z per revolution */
	     gleDouble startXform[2][3], /* starting contour affine xform */
	     gleDouble dXformdTheta[2][3], /* tangent change xform per revoln */
	     gleDouble startTheta,      /* start angle in x-y plane */
	     gleDouble sweepTheta);     /* degrees to spiral around */

/* similar to spiral, except contour is a circle */
extern void gleHelicoid (gleDouble rToroid, /* circle contour (torus) radius */
	     gleDouble startRadius,     /* spiral starts in x-y plane */
	     gleDouble drdTheta,        /* change in radius per revolution */
	     gleDouble startZ,          /* starting z value */
	     gleDouble dzdTheta,        /* change in Z per revolution */
	     gleDouble startXform[2][3], /* starting contour affine xform */
	     gleDouble dXformdTheta[2][3], /* tangent change xform per revoln */
	     gleDouble startTheta,      /* start angle in x-y plane */
	     gleDouble sweepTheta);     /* degrees to spiral around */

/* similar to lathe, except contour is a circle */
extern void gleToroid (gleDouble rToroid, /* circle contour (torus) radius */
	     gleDouble startRadius,     /* spiral starts in x-y plane */
	     gleDouble drdTheta,        /* change in radius per revolution */
	     gleDouble startZ,          /* starting z value */
	     gleDouble dzdTheta,        /* change in Z per revolution */
	     gleDouble startXform[2][3], /* starting contour affine xform */
	     gleDouble dXformdTheta[2][3], /* tangent change xform per revoln */
	     gleDouble startTheta,      /* start angle in x-y plane */
	     gleDouble sweepTheta);     /* degrees to spiral around */

/* draws a screw shape */
extern void gleScrew (int ncp,          /* number of contour points */
	     gleDouble contour[][2],    /* 2D contour */
	     gleDouble cont_normal[][2], /* 2D contour normals */
	     gleDouble up[3],           /* up vector for contour */
	     gleDouble startz,          /* start of segment */
	     gleDouble endz,            /* end of segment */
	     gleDouble twist);          /* number of rotations */

extern void gleTextureMode (int mode);
};

#endif /* STORMPROTOS_GLE_SPROTOS_H */

